/*	A binary file I/O class

	Desc:	A binary file class that uses <<,>> rather than write/read to perform file stream I/O

	Author:		Sobeit Void <darkaurora@yahoo.com>
	Revision:	1.0
	Date:		20/11/2000

	The source code is provided as it is, without any warranty or liability. You are free to modify the
	code for your own use.The author is not liable for any damages resulting from the usage of the code.

	Output Summary:

		simple types like int, float,..	- use <<
		string and buffers				- use << as long as they are terminated. Terminate them by ends
		stl strings						- use <<. Terminate them be ends
		arrays and pointers				- use write but you can loop through each element and use <<
		class objects					- use a bofstream friend operator<< just like the std streams

	Input Summary:

		simple types like int, float,..	- use >>
		arrays and pointers				- use read but can use a loop and use >>
		string and buffers				- use getline, not >>. You have been warned!
		stl strings						- use >> (if deliminter is '\0') or else use getline
		class objects					- use a bifstream friend operator >> just like the std streams

	Formatting manipulators (Non standard)

		_								- inserts a space if text mode. Does nothing in binary mode
		t_								- inserts a tab if text mode. Does nothing in binary mode
		endl							- inserts a newline if text mode. Does nothing in binary mode
		ends							- inserts a '\0' char. Inserts a newline if text mode

	Avoid using these manipulators in text mode

		putback()
		peek()
		ignore()
		gcount()
		seeking operators like seekp(), seekg(), tellp() etc. unless the seek is to beginning or end only
*/

#pragma once

#include <fstream>
#include <string>
#include <limits>

//namespace V
//{

//////////////////////////////

//--------------------------------------------------------------------------------------
//	Auxilary class to provide common functions in bofstream and bifstream class and is
//	meant to be instantiated
//--------------------------------------------------------------------------------------

class	bfstream_aux
{
	typedef	std::ios					ios;
	typedef	std::fstream				fstream;
	typedef	std::string					string;
	typedef	std::ios::open_mode			open_mode;
	typedef	std::ios_base::fmtflags		fmtflags;
	typedef	std::ios_base::iostate		iostate;

public:

	bool is_binary()	const	{	return bBinary_;		}		// returns true if binary mode

	void close()
	{
		if (fs_.is_open())
			fs_.close();
	}

	// Operator for error testing
	//////////////////////////////

	bool operator!()	const	{	return !fs_;			}

	bool is_open()		const	{	return fs_.is_open();	}

	bool good()			const	{	return fs_.good();		}

	bool eof()			const	{	return fs_.eof();		}

	bool fail()			const	{	return fs_.fail();		}

	bool bad()			const	{	return fs_.bad();		}

	iostate	rdstate()	const	{	return fs_.rdstate();	}

	void clear(const iostate state = ios::goodbit)	{	fs_.clear(state);		}

	void setstate(const iostate state)				{	fs_.setstate(state);	}

	// Formatting specifiers
	/////////////////////////

	fmtflags	setf(const fmtflags flags)						{	return fs_.setf(flags);			}
	fmtflags	setf(const fmtflags flags, const fmtflags mask)	{	return fs_.setf(flags, mask);	}

	void		unsetf(const fmtflags mask)		{	fs_.unsetf(mask);			}

	fmtflags	flags()					const	{	return fs_.flags();			}
	fmtflags	flags(const fmtflags flags)		{	return fs_.flags(flags);	}

protected:

	// Constructors made protected as this class is not meant to be instantiated

	bfstream_aux(const string &name, const open_mode mode)
		: fs_(name.c_str(), Set_Flag(mode))		{}

	bfstream_aux() : fs_(), bBinary_(false)		{}

	//@@ Closes and reopen the file stream

	void open_stream(const string &name, const open_mode mode)
	{
		close();
		fs_.open(name.c_str(), Set_Flag(mode));
	}

private:

	//@@ Sets internal binary flag on/off based on creation flags
	//@@ Remarks: To be called at during opening of the file stream

	open_mode	Set_Flag(const open_mode mode)
	{
		bBinary_ = (ios::binary & mode) ? true : false;	// if binary mode is specified in the passed flags, set flag on

		return mode;									// return the original passed parameters
	}

	// =================== data ==================

protected:

	bool		bBinary_;		// true if IO is in binary mode

	fstream		fs_;			// the file stream
};

/////////////

//------------------------------------------------------------------------------------------
//	The input class
//------------------------------------------------------------------------------------------

class	bifstream :	virtual public	bfstream_aux
{
	typedef	std::ios					ios;
	typedef std::numeric_limits<int>	limit_int;
	typedef std::streamsize				streamsize;
	typedef	std::string					string;
	typedef	std::ios::open_mode			open_mode;
	typedef std::ios::pos_type			pos_type;
	typedef std::ios_base::seekdir		seekdir;

public:

	// constructors
	////////////////

	bifstream(const string &name, const open_mode mode = 0)
		: bfstream_aux(name, mode | ios::in)	{}				// text and in mode implied

	bifstream() : bfstream_aux()		{};

	void open(const string &name, const open_mode mode = 0)		// text mode always implied
	{
		open_stream(name, mode | ios::in);						// in mode always implied
	}

	// input operators
	///////////////////

	template <typename T>
		bifstream& operator>>(T& data)
	{
		if (bBinary_)								// if binary mode
			fs_.read((char *)&data, sizeof(data));
		else										// text mode
			fs_ >> data;

		return *this;
	}

	// template specialization for std::string type

	bifstream& operator>>(string &s)
	{
		text_skip_to_newline();

		std::getline(fs_, s, '\0');		// the string must be deliminted by a ends
		return *this;
	}

	template <typename T>
		bifstream& read(T& data, const streamsize size)
	{
		text_skip_to_newline();

		fs_.read((char *)data, size);
		return *this;
	}

	bifstream&	getline(char *buffer, const int count, const char delim = '\0')
	{
		text_skip_to_newline();

		fs_.getline(buffer, count, delim);
		return *this;
	}

	bifstream&	getline(string &str, const char delim = '\0')
	{
		text_skip_to_newline();

		std::getline(fs_, str, delim);
		return *this;
	}

	//@@ Remarks: Avoid usage in text mode

	streamsize	gcount()	const	{	return fs_.gcount();	}

	//@@ Remarks: Avoid usage in text mode

	bifstream& ignore(const streamsize count = 1, const int delim = std::char_traits<char>::eof())
	{
		fs_.ignore(count, delim);
		return *this;
	}

	//@@ Remarks: Avoid usage in text mode

	int	peek()			{	return fs_.peek();	}

	//@@ Remarks: Avoid usage in text mode

	bifstream&	putback(char c)		{	fs_.putback(c);		return *this;	}

	// seeking operators - exercise caution when using in text mode
	///////////////////////////////////////////////////////////////

	bifstream&	seekg(const pos_type pos)						{	fs_.seekg(pos);			return *this;	}
	bifstream&	seekg(const pos_type pos, const seekdir dir)	{	fs_.seekg(pos, dir);	return *this;	}

	pos_type	tellg()		{	return fs_.tellg();		}

private:

	//@@ Moves the file position to a newline in if the file is in text mode. Function does nothing in binary mode

	void text_skip_to_newline()
	{
		if (!bBinary_)
			fs_.ignore(limit_int::max(),'\n');		// ignore everything until a newline is met
	}

};

///////////////////////

//--------------------------------------------------------------------------------------------
//	The output class
//--------------------------------------------------------------------------------------------

class	bofstream	:	virtual public bfstream_aux
{
	typedef	std::ios					ios;
	typedef std::streamsize				streamsize;
	typedef	std::string					string;
	typedef	std::ios::open_mode			open_mode;
	typedef std::ios::pos_type			pos_type;
	typedef std::ios_base				ios_base;
	typedef std::ios_base::seekdir		seekdir;

public:

	// constructors
	////////////////

	bofstream(const string &name, const open_mode mode = std::ios::trunc)
		: bfstream_aux(name, mode | ios::out)		{}				// trunc and out mode implied

	bofstream() : bfstream_aux()					{};

	void open(const string &name, const open_mode mode = ios::trunc)	// trunc mode implied
	{
		open_stream(name, mode | ios::out);								// out mode always implied
	}

	// output operators
	////////////////////

	// output for basic types

	template <typename T>
		bofstream& operator<<(const T& data)
	{
		if (bBinary_)										// if binary mode
			fs_.write((const char *)&data, sizeof(data));
		else												// text mode
			fs_ << data;

		return *this;
	}

	// template specialization for char buffers as sizeof() doesn't work with string types
	// traps		string array			- char buffer[20]
	//				pointer string			- char *p = new char[20]
	//				const string			- "Const string"
	// parameter cannot be const to trap pointer string

	bofstream& operator<<(char *data)
	{
		fs_.write(data, strlen(data));		// write out the length of the string
		return *this;
	}

	// template specialization for const string pointers and const string const pointers
	// traps		const string pointers		- const char *p = "Const string"
	//				const string const pointers	- const char * const p = "Const string"

	bofstream& operator<<(const char *data)
	{
		fs_.write(data, strlen(data));		// write out the length of the string
		return *this;
	}

	// template specialization for std::string type

	bofstream& operator<<(const string &s)
	{
		fs_.write(s.c_str(), s.length());	// write out the length of the string
		return *this;
	}

	template <typename T>
		bofstream&	write(const T &data, const streamsize size)
	{
		fs_.write((const char *)data, size);

		textput('\n');					// if text mode, place a newline to make a new block

		return *this;
	}

	bofstream&	put(const char c)		{	fs_.put(c);		return *this;	}

	//@@ Outputs the passed char if the file is in text mode. Function does nothing in binary mode.

	void	textput(const char c)
	{
		if (!bBinary_)			// if not binary mode, then do the output
			fs_.put(c);
	}

	// Seeking operators - exercise caution when using in text mode
	///////////////////////////////////////////////////////////////

	bofstream&	seekp(const pos_type pos)						{	fs_.seekp(pos);			return *this;	}
	bofstream&	seekp(const pos_type pos, const seekdir dir)	{	fs_.seekp(pos, dir);	return *this;	}

	pos_type	tellp()			{	return fs_.tellp();		}

	// manipulators
	////////////////

	// Function pointer for passing in endl, ends, flush

	bofstream& operator<<(void (*fp)(bofstream&))	{	(*fp)(*this);	return *this;	}

	bofstream&	flush()								{	fs_.flush();	return *this;	}

	streamsize	precision()					const	{	return fs_.precision();			}
	streamsize	precision(const streamsize size)	{	return fs_.precision(size);		}

	streamsize	width()						const	{	return fs_.width();				}
	streamsize	width(const streamsize size)		{	return fs_.width(size);			}
};

///////////////////

// manipulators
///////////////

inline void flush(bofstream& s)		{	s.flush();		}

inline void ends(bofstream& s)		{	s.put('\0');	s.textput('\n');	}

// text mode manipulators - only outputs the char in text mode, does nothing in binary mode

inline void endl(bofstream& s)		{	s.textput('\n');	}		// puts a newline in text mode only, no flush, unlike standard streams

inline void _(bofstream& s)			{	s.textput(' ');		}		// puts a space in text mode only

inline void t_(bofstream& s)		{	s.textput('\t');	}		// puts a tab in text mode only

////////////////////

//----------------------------------------------
// auxilary class for parameterized manipulators
//----------------------------------------------

template <typename T>
class Bfmanip
{
	typedef void(*FP)(bofstream&, T);		// function pointer prototype

public:
	Bfmanip(FP fp , T arg) : fp_(fp), arg_(arg)	{};

	friend bofstream& operator<<(bofstream& os, const Bfmanip &manip)
	{
		(*manip.fp_)(os, manip.arg_);
		return os;
	}

	friend bifstream& operator>>(bifstream& is, const Bfmanip &manip)
	{
		(*manip.fp_)(is, manip.arg_);
		return is;
	}

	// =========== data ===============

private:

	FP		fp_;		// function pointer
	T		arg_;		// arg passed to function pointer
};

// auxiliary functions for parameterized manipulators
/////////////////////////////////////////////////////

inline void setiosflags__(bofstream& os, std::ios_base::fmtflags m)		{	os.setf(m);					}
inline void resetiosflags__(bofstream& os, std::ios_base::fmtflags m)	{	os.setf(os.flags(), m);		}

inline void setprecision__(bofstream& os, const int i)		{	os.precision(i);	}
inline void setw__(bofstream& os, const int i)				{	os.width(i);		}

// parameterized manipulators that affect text mode only
/////////////////////////////////////////////////////////

inline Bfmanip<std::ios_base::fmtflags> setiosflags(std::ios_base::fmtflags mask)	
{
	return Bfmanip<std::ios_base::fmtflags>(setiosflags__, mask);
}

inline Bfmanip<std::ios_base::fmtflags> resetiosflags(std::ios_base::fmtflags mask)
{
	return Bfmanip<std::ios_base::fmtflags>(resetiosflags__, mask);
}

inline Bfmanip<const int>	setprecision(const int prec)	{	return Bfmanip<const int>(setprecision__, prec);	}

inline Bfmanip<const int>	setw(const int width)			{	return Bfmanip<const int>(setw__, width);			}

///////////////////////////////

//};	// end namespace